



luigi --help
IF %ERRORLEVEL% NEQ 0 exit 1
luigid --help
IF %ERRORLEVEL% NEQ 0 exit 1
luigi-grep --help
IF %ERRORLEVEL% NEQ 0 exit 1
luigi-deps --help
IF %ERRORLEVEL% NEQ 0 exit 1
luigi-deps-tree --help
IF %ERRORLEVEL% NEQ 0 exit 1
exit 0
